#include <SPI.h>
#include <SD.h>
#include <Adafruit_PCD8544.h>

#define SH_INPUT 7
#define SH_CLOCK 5
#define OR_CLOCK 6
#define MOSI 3
#define SCK 6
#define SS 5
#define MISO 4
#define GREEN_LED 0
#define WHITE_LED 1
#define BLUE_LED 2
#define Buzzer 3

// Nokia 5110 LCD
Adafruit_PCD8544 display = Adafruit_PCD8544(8,9,10,11,12);

// ports for joystick
int VRx = A0;
int VRy = A1;
int SW = 2;

// data from joystick
int xPosition = 0;
int yPosition = 0;
int SW_state = 1;
int mapX = 0;
int mapY = 0;

// these are some variables used in the program flow
char rotate = 1;
float playerX = 20;
float playerY = 80;
int deltaX, deltaY;
long deltaTime = 0;
char currentLevel = 1;
char missedObjects = 0;
int totalMissedObjects = 0;
int explodedObjects;
int score = 0;
char nrLives = 3;
bool gameOver = false;
byte led = 0;
int MAXScore = 0;
File myFile;

// structure of an object
typedef struct obj {
        float x;
        float y;
        long t;
        float objSpeed;
        bool active;
} object;

// structure of bullet
typedef struct bullet {
        float x;
        float y;
        float bulletSpeed;
        bool active;
} _bullet;

object objects[6];
_bullet bullet[6];

void setup()   {

    // set the joystick input pins
    pinMode(VRx, INPUT);
    pinMode(VRy, INPUT);
    pinMode(SW, INPUT_PULLUP);
  
    Serial.begin(115200);
    display.begin();
    display.setContrast(60);

    // set the ports for ShiftRegister
    pinMode(SH_INPUT, OUTPUT);
    pinMode(SH_CLOCK, OUTPUT); 
    pinMode(OR_CLOCK, OUTPUT);

    turnOnLeds();
    byte x = 0;
    bitSet(x, 5);
    updateShiftRegister(x);
    if (!SD.begin()) {
      Serial.println("initialization failed!");
      while(1);
    }
    Serial.println("initialization done.");
   readMAXScore();
    
    display.setRotation(3);
    display.clearDisplay();
    
    level1_setup();
    drawPlayer(playerX, playerY); 
    
}

// read max score from SDcard
void readMAXScore()
{
  myFile = SD.open("file.txt");
  if (myFile) {
    Serial.println("file.txt:");

    if (myFile.available()) {
      MAXScore = myFile.read();
    }
  
    myFile.close();
   } else {
    Serial.println("error opening file.txt");
  }
}

// write a new max score record
void writeMAXScore()
{
  myFile = SD.open("file.txt", FILE_WRITE);

  if (myFile) {
    myFile.print(MAXScore);
    myFile.close();
  } else {
    Serial.println("error opening file.txt");
  }
}

void updateShiftRegister(byte updateLed)
{
  digitalWrite(OR_CLOCK, LOW);
  shiftOut(SH_INPUT, SH_CLOCK, MSBFIRST, updateLed);
  digitalWrite(OR_CLOCK, HIGH);
}

void turnOnLeds()
{
  for (int i = 0; i < 3; i++) {
    bitSet(led, i);
    updateShiftRegister(led);
  }
  
}

void turnOffLeds(int Led, bool leds)
{
  if (leds) {
    for (int i = 0; i < 3; i++) {
      bitClear(led, i);
      updateShiftRegister(led);
    }
  } else {
      bitClear(led, Led);
      updateShiftRegister(led);
    }

}

// this is the jingle when you pass the current level
void startJingle()
{
  delay(500);
  tone(Buzzer, 1800, 100);
  delay(100);
  tone(Buzzer, 2000, 100);
  delay(100);
  tone(Buzzer, 2200, 100);
  delay(1500);
}

// this is the display show when you pass the current level
void setDisplay(int level)
{
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(BLACK);
  display.setCursor(5,15);
  display.print("Level ");
  display.println(level);
  display.display();
}

// reset objects and bullets data at the beginning of each level
void setLevel(int level, int deltatime, float objSpeed)
{
  totalMissedObjects += missedObjects;
  missedObjects = 0;
  deltaTime = deltatime;
  for (int i = 0; i < 6; i++) {
    objects[i].x = random(2, 43);
    objects[i].y = 2;
    objects[i].t = (level == 1) ? (i + 1) * 4000 : (level == 2) ? millis() + (i + 1) * 3000 : millis() + (i + 1) * 2000; 
    objects[i].objSpeed = objSpeed;
    objects[i].active = false;

    bullet[i].x = playerX + 2;
    bullet[i].y = playerY - 9;
    bullet[i].bulletSpeed = 4;
    bullet[i].active = false;  
  }
}

void level1_setup()
{
  currentLevel = 1;
  setDisplay(currentLevel);
  delay(2500);
  setLevel(currentLevel, 4000, 2);
}

void level2_setup()
{
  turnOnLeds();
  currentLevel = 2;
  setDisplay(currentLevel);
  startJingle();
  display.clearDisplay();
  setLevel(currentLevel, 3000, 2.5);
}

void level3_setup()
{
  turnOnLeds();
  currentLevel = 3;
  setDisplay(currentLevel); 
  startJingle();
  setLevel(currentLevel, 2000, 3);
}

void drawPlayer(float x, float y)
{
    display.drawTriangle(x - 2.5, y, x + 2.5, y, (2 * (x - 2.5) + 4.9) / 2, y - 10, BLACK);
    display.drawTriangle(x + 2.5, y, x + 7.5, y, (2 * (x + 2.5) + 4.9) / 2, y - 10, BLACK);
    display.display(); 
}

void drawBullet()
{
  for (int i = 0; i < 6; i++) {
    if (bullet[i].active) {
        if (bullet[i].y + bullet[i].bulletSpeed >= 0 && bullet[i].y + bullet[i].bulletSpeed <= 84 && !bulletCollision(i))
            bullet[i].y -= bullet[i].bulletSpeed;
        else
            bullet[i].active = false;
            
        display.drawCircle(bullet[i].x, bullet[i].y, 2, BLACK);
        display.fillCircle(bullet[i].x, bullet[i].y, 2, BLACK);
        display.display();
    } 
  }

}

// generate the objects considering bullet - objects, objects - player and objects - floor collisions
void drawObject()
{
  for (int i = 0; i < 6; i++) {
    long elapsedTime = millis();

    if (!objects[i].active && elapsedTime - objects[i].t >= deltaTime) {
        objects[i].x = random(1, 43);
        objects[i].y = 0;
        objects[i].active = true;
    } else if (objects[i].active){
        display.drawRect(objects[i].x, objects[i].y, 5, 5, BLACK);
        display.fillRect(objects[i].x, objects[i].y, 5, 5, BLACK);
        display.display(); 
        if (objects[i].y + objects[i].objSpeed < 80 && !(abs(playerX - objects[i].x) <= 6 && abs(playerY - objects[i].y) <= 4))
            objects[i].y += objects[i].objSpeed;
        else {
            objects[i].y = 0;
            objects[i].active = false;
            objects[i].t = millis();
            totalMissedObjects++;
            missedObjects++;
        }
     }    
   }
   
}

void movePlayer(float *x, float *y, int deltaX, int deltaY)
{ 
  if (*x + deltaX > 0 && *x + deltaX < 41)
   *x += deltaX;

  if (*y + deltaY > 4 && *y + deltaY < 84)
   *y += deltaY;

  for (int i = 0; i < 5; i++) {
    if (!bullet[i].active) {
        bullet[i].x = *x + 2;
        bullet[i].y = *y - 9;
    } 
  }

}

// check bullet - objects collisions
bool bulletCollision(int i)
{
  for (int j = 0; j < 6; j++) {
    if (objects[j].active && abs(bullet[i].x - objects[j].x - 2) <= 5 && abs(bullet[i].y - objects[j].y) <= 4) {
      objects[j].y = 0;
      objects[j].active = false;
      objects[j].t = millis();
      explodedObjects++;    
      tone(Buzzer, 2000, 100);
      return true;  
    }
  }

  return false;
  
}

// The player has 3 lives per level, I update the number of his lives
// considering missed objects (those objects that were not exploded) and
// the player's collisions with the objects
void checkPlayerLives()
{
  if (currentLevel == 1) {
    if (missedObjects == 1) {
        turnOffLeds(GREEN_LED, false);
    } else if (missedObjects == 2) {
        turnOffLeds(WHITE_LED, false);
    } else if (missedObjects == 3) {
        turnOffLeds(BLUE_LED, false);
        gameOver = true;
    }  
    
  } else if (currentLevel == 2) {
     if (missedObjects == 2) {
        turnOffLeds(GREEN_LED, false);
    } else if (missedObjects == 4) {
        turnOffLeds(WHITE_LED, false);
    } else if (missedObjects == 6) {
        turnOffLeds(BLUE_LED, false);
        gameOver = true;
    }
  
  } else {
     if (missedObjects == 4) {
        turnOffLeds(GREEN_LED, false);
    } else if (missedObjects == 8) {
        turnOffLeds(WHITE_LED, false);
    } else if (missedObjects == 12) {
        turnOffLeds(BLUE_LED, false);
        gameOver = true;
    }
  }
        
}

// compute the score considering a magic formula
void computeScore()
{
  score = 50 * (explodedObjects - 2) - 15 * totalMissedObjects;  
}

bool checkGameOver()
{
    if (gameOver == true) {
      computeScore();
      if (score > MAXScore) {
        writeMAXScore();
        display.clearDisplay();
        display.setTextSize(1);
        display.setTextColor(BLACK);
        display.setCursor(0,15);
        display.println("GAMEOVER");
        display.setCursor(15,35);
        display.println("Your ");
        display.setCursor(10,45);
        display.println("Score: ");
        display.setCursor(15,55);
        display.println(score);
      } else {
        display.setCursor(0,15);  
        display.println("GAMEOVER");
        display.setCursor(0,35);
        display.println(" You set ");
        display.setCursor(5,45);
        display.println(" a new  record!");
        display.setCursor(15,65);
        display.println(score);
      }
      
      display.display();
      delay(500);
      tone(Buzzer, 1800, 1000);
      delay(500);
      tone(Buzzer, 1500, 1000);
      delay(500);
      tone(Buzzer, 1200, 2000);
      delay(1500);
      while(1) {}
    }

}

void loop() {

    // map the read data from joystick for better precision
    xPosition = analogRead(VRx);
    yPosition = analogRead(VRy);
    SW_state = digitalRead(SW);
    mapX = map(xPosition, 0, 670, -9, 10);
    mapY = map(yPosition, 0, 670, -9, 10);
    

    if (explodedObjects == 4){ 
      level2_setup();
      explodedObjects++;
    }
    
    if (explodedObjects == 14) {
     level3_setup();
     explodedObjects++;
    }

    int ok = 0;
    if (SW_state == 1) {
      for (int i = 0; i < 6 && ok == 0; i++) {
        if (!bullet[i].active) {
          bullet[i].active = true;
          ok = 1;
        }
      }
    }
    
    movePlayer(&playerX, &playerY, mapX, mapY);
    display.clearDisplay();
    drawBullet();
    drawObject();
    drawPlayer(playerX, playerY);
    checkPlayerLives();
    checkGameOver();
  
    delay(80);
  
}
